//=================================================================================
//
//					Perspective Fast Marching Method (Kernel)
//					Date: 30-04-2007
//					By Chi-kin CHOW
//					Version 1.0
//
//=================================================================================

#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include<malloc.h>
#include"..\ACC_Math_Kernel\ACC_Math_Kernel.H"
#include"..\SFS_Structure_Kernel\SFS_Structure_Kernel.H"
#include"..\Perspective_FMM_Frontal_Kernel\Perspective_FMM_Frontal_Kernel.H"

void Perspective_FMM_Frontal_Initialization(SFS_Element *cur_Psp_FMM)
{
	int i, j;
	int row_index, col_index;

	for(i=0; i< (*cur_Psp_FMM).no_Row; ++i)
		for(j=0; j< (*cur_Psp_FMM).no_Col; ++j)
		{
			if((*cur_Psp_FMM).IntensityMap[i][j] == INF)
				(*cur_Psp_FMM).OrderMap[i][j] = -1;
			else
				(*cur_Psp_FMM).OrderMap[i][j] = 2;
			(*cur_Psp_FMM).estDepthMap[i][j] = INF;
		}

	for(i=0; i< (*cur_Psp_FMM).no_SingularPoint; ++i)
	{
		row_index = (int) (*cur_Psp_FMM).SingularPoint_Set[i][0];
		col_index = (int) (*cur_Psp_FMM).SingularPoint_Set[i][1];

		(*cur_Psp_FMM).OrderMap[row_index][col_index] = 0;
		(*cur_Psp_FMM).estDepthMap[row_index][col_index] = (*cur_Psp_FMM).SingularPoint_Set[i][2];

		if(row_index > 0 && (*cur_Psp_FMM).OrderMap[row_index - 1][col_index] == 2)
			(*cur_Psp_FMM).OrderMap[row_index - 1][col_index] = 1;

		if(row_index < (*cur_Psp_FMM).no_Row - 1 && (*cur_Psp_FMM).OrderMap[row_index + 1][col_index] == 2)
			(*cur_Psp_FMM).OrderMap[row_index + 1][col_index] = 1;

		if(col_index > 0 && (*cur_Psp_FMM).OrderMap[row_index][col_index - 1] == 2)
			(*cur_Psp_FMM).OrderMap[row_index][col_index - 1] = 1;

		if(col_index < (*cur_Psp_FMM).no_Col - 1 && (*cur_Psp_FMM).OrderMap[row_index][col_index + 1] == 2)
			(*cur_Psp_FMM).OrderMap[row_index][col_index + 1] = 1;
	}

	return;
}

void Perspective_FMM_Frontal_Kernel(SFS_Element *cur_Psp_FMM)
{
	int i, j;
	int no_consider_pixel, no_accept_pixel;
	int cur_no_consider_pixel, cur_consider_pixel_list[4][2];
	int cur_consider_row_index, cur_consider_col_index;
	int cur_accept_row_index, cur_accept_col_index;
	double min_consider_depth;

	//============ Initialize the depth of considered pixel(s)
	for(i=0; i< (*cur_Psp_FMM).no_Row; ++i)
		for(j=0; j< (*cur_Psp_FMM).no_Col; ++j)
			if((*cur_Psp_FMM).OrderMap[i][j] == 1)
			{
				Perspective_FMM_Frontal_Depth_CC(i, j, cur_Psp_FMM);
				if((*cur_Psp_FMM).estDepthMap[i][j] == INF)
					(*cur_Psp_FMM).OrderMap[i][j] = -1;
			}

	no_accept_pixel = 0;
	do {
		no_consider_pixel = 0;
		for(i=0; i< (*cur_Psp_FMM).no_Row; ++i)
			for(j=0; j< (*cur_Psp_FMM).no_Col; ++j)
				if ((*cur_Psp_FMM).OrderMap[i][j] == 1)
				{
					if(no_consider_pixel == 0 || min_consider_depth > (*cur_Psp_FMM).estDepthMap[i][j])
					{
						min_consider_depth = (*cur_Psp_FMM).estDepthMap[i][j];
						cur_accept_row_index = i;
						cur_accept_col_index = j;
					}
					++no_consider_pixel;
				}

		if(no_consider_pixel > 0)
		{
			(*cur_Psp_FMM).OrderMap[cur_accept_row_index][cur_accept_col_index] = 0;
			++no_accept_pixel;

			//============ Update the OrderMap and the depths of considered pixel(s)
			cur_no_consider_pixel = 0;
			if(cur_accept_row_index > 0 &&
			   ((*cur_Psp_FMM).OrderMap[cur_accept_row_index - 1][cur_accept_col_index] == 1 ||
			    (*cur_Psp_FMM).OrderMap[cur_accept_row_index - 1][cur_accept_col_index] == 2))
			{
				cur_consider_pixel_list[cur_no_consider_pixel][0] = cur_accept_row_index - 1;
				cur_consider_pixel_list[cur_no_consider_pixel][1] = cur_accept_col_index;
				++cur_no_consider_pixel;
			}

			if(cur_accept_row_index < (*cur_Psp_FMM).no_Row - 1 &&
			   ((*cur_Psp_FMM).OrderMap[cur_accept_row_index + 1][cur_accept_col_index] == 1 ||
			    (*cur_Psp_FMM).OrderMap[cur_accept_row_index + 1][cur_accept_col_index] == 2))
			{
				cur_consider_pixel_list[cur_no_consider_pixel][0] = cur_accept_row_index + 1;
				cur_consider_pixel_list[cur_no_consider_pixel][1] = cur_accept_col_index;
				++cur_no_consider_pixel;
			}

			if(cur_accept_col_index > 0 &&
			   ((*cur_Psp_FMM).OrderMap[cur_accept_row_index][cur_accept_col_index - 1] == 1 ||
			    (*cur_Psp_FMM).OrderMap[cur_accept_row_index][cur_accept_col_index - 1] == 2))
			{
				cur_consider_pixel_list[cur_no_consider_pixel][0] = cur_accept_row_index;
				cur_consider_pixel_list[cur_no_consider_pixel][1] = cur_accept_col_index - 1;
				++cur_no_consider_pixel;
			}

			if(cur_accept_col_index < (*cur_Psp_FMM).no_Col - 1 && 
			   ((*cur_Psp_FMM).OrderMap[cur_accept_row_index][cur_accept_col_index + 1] == 1 ||
			    (*cur_Psp_FMM).OrderMap[cur_accept_row_index][cur_accept_col_index + 1] == 2))
			{
				cur_consider_pixel_list[cur_no_consider_pixel][0] = cur_accept_row_index;
				cur_consider_pixel_list[cur_no_consider_pixel][1] = cur_accept_col_index + 1;
				++cur_no_consider_pixel;
			}

			for(i=0; i< cur_no_consider_pixel; ++i)
			{
				cur_consider_row_index = cur_consider_pixel_list[i][0];
				cur_consider_col_index = cur_consider_pixel_list[i][1];
				(*cur_Psp_FMM).OrderMap[cur_consider_row_index][cur_consider_col_index] = 1;

				Perspective_FMM_Frontal_Depth_CC(cur_consider_row_index, cur_consider_col_index, cur_Psp_FMM);
				if((*cur_Psp_FMM).estDepthMap[cur_consider_row_index][cur_consider_col_index] == INF)
					(*cur_Psp_FMM).OrderMap[cur_consider_row_index][cur_consider_col_index] = -1;
			}
		}

		if(no_accept_pixel % 200 == 0)
			printf("No. of accepted pixel = %d, No. of considered pixel = %d\n", no_accept_pixel, no_consider_pixel);

	} while(no_consider_pixel > 0);
//	} while(no_accept_pixel < 1200 && no_consider_pixel > 0);

	return;
}

void Perspective_FMM_Frontal_Depth_CC(int row_index_0, int col_index_0, SFS_Element *cur_Psp_FMM)
{
	int mode;
	int row_index_1, row_index_2;
	int col_index_1, col_index_2;
	double Intensity, Intensity2, Focal_Length, If_2;
	double u0, u1, u2;
	double v0, v1, v2;
	double D0_neg, D0_pos, D0_tmp;
	double D1_neg, D1_pos, D1;
	double D2_neg, D2_pos, D2;
	double min_D1_D2, max_D1_D2;
	double A1, A2, A3;
	double B1, B2, B3;
	double DA, DB, DC, DD;

	Focal_Length = (*cur_Psp_FMM).Focal_Length;

	u0 = (*cur_Psp_FMM).Grid_X[row_index_0][col_index_0];
	v0 = (*cur_Psp_FMM).Grid_Y[row_index_0][col_index_0];
	Intensity = (*cur_Psp_FMM).IntensityMap[row_index_0][col_index_0];
	Intensity2 = pow(Intensity, 2);
	If_2 = pow(Focal_Length * Intensity, 2);

	if (Intensity == INF)
	{
		(*cur_Psp_FMM).estDepthMap[row_index_0][col_index_0] = INF;
		return;
	}

	//============ P2
	if(row_index_0 == 0 || (*cur_Psp_FMM).OrderMap[row_index_0 - 1][col_index_0] != 0)
		D2_neg = INF;
	else
		D2_neg = (*cur_Psp_FMM).estDepthMap[row_index_0 - 1][col_index_0];

	if(row_index_0 == (*cur_Psp_FMM).no_Row - 1 || (*cur_Psp_FMM).OrderMap[row_index_0 + 1][col_index_0] != 0)
		D2_pos = INF;
	else
		D2_pos = (*cur_Psp_FMM).estDepthMap[row_index_0 + 1][col_index_0];

	if(D2_neg == INF && D2_pos == INF)
	{
		if(row_index_0 == (*cur_Psp_FMM).no_Row - 1)
			row_index_2 = row_index_0 - 1;
		else
			row_index_2 = row_index_0 + 1;
		col_index_2 = col_index_0;
		D2 = INF;
	}
	else if(D2_neg < D2_pos)
	{
		row_index_2 = row_index_0 - 1;
		col_index_2 = col_index_0;
		D2 = (*cur_Psp_FMM).estDepthMap[row_index_2][col_index_2];
	}
	else
	{
		row_index_2 = row_index_0 + 1;
		col_index_2 = col_index_0;
		D2 = (*cur_Psp_FMM).estDepthMap[row_index_2][col_index_2];
	}
	u2 = (*cur_Psp_FMM).Grid_X[row_index_2][col_index_2];
	v2 = (*cur_Psp_FMM).Grid_Y[row_index_2][col_index_2];

	//============ P1
	if(col_index_0 == 0 || (*cur_Psp_FMM).OrderMap[row_index_0][col_index_0 - 1] != 0)
		D1_neg = INF;
	else
		D1_neg = (*cur_Psp_FMM).estDepthMap[row_index_0][col_index_0 - 1];

	if(col_index_0 == (*cur_Psp_FMM).no_Col - 1 || (*cur_Psp_FMM).OrderMap[row_index_0][col_index_0 + 1] != 0)
		D1_pos = INF;
	else
		D1_pos = (*cur_Psp_FMM).estDepthMap[row_index_0][col_index_0 + 1];

	if(D1_neg == INF && D1_pos == INF)
	{
		row_index_1 = row_index_0;
		if(col_index_0 == (*cur_Psp_FMM).no_Col - 1)
			col_index_1 = col_index_0 - 1;
		else
			col_index_1 = col_index_0 + 1;
		D1 = INF;
	}
	else if(D1_neg < D1_pos)
	{
		row_index_1 = row_index_0;
		col_index_1 = col_index_0 - 1;
		D1 = (*cur_Psp_FMM).estDepthMap[row_index_1][col_index_1];
	}
	else
	{
		row_index_1 = row_index_0;
		col_index_1 = col_index_0 + 1;
		D1 = (*cur_Psp_FMM).estDepthMap[row_index_1][col_index_1];
	}
	u1 = (*cur_Psp_FMM).Grid_X[row_index_1][col_index_1];
	v1 = (*cur_Psp_FMM).Grid_Y[row_index_1][col_index_1];

	if(D1 == INF)
		min_D1_D2 = max_D1_D2 = D2;
	else if(D2 == INF)
		min_D1_D2 = max_D1_D2 = D1;
	else if(D1 < D2)
	{
		min_D1_D2 = D1;
		max_D1_D2 = D2;
	}
	else
	{
		min_D1_D2 = D2;
		max_D1_D2 = D1;
	}

	(*cur_Psp_FMM).estDepthMap[row_index_0][col_index_0] = INF;

	if(D1 != INF && D2 != INF)
	{
		//============ U & V
		mode = 0;

		A1 = D1*(v0-v1) - D2*(v0-v2);
		A2 = D2*(u0-u2) - D1*(u0-u1);
		A3 = (D2*(u2*v0-u0*v2) + D1*(u0*v1-u1*v0))/Focal_Length;
    
		B1 = D1*D2*(v1-v2);
		B2 = D1*D2*(u2-u1);
		B3 = D1*D2*(v2*u1-u2*v1)/Focal_Length;

		DA = Intensity2*(pow(A1, 2) + pow(A2, 2) + pow(A3, 2)) - pow(A3, 2);
		DB = 2.0*Intensity2*(A1*B1 + A2*B2 + A3*B3) - 2.0 * A3*B3;
		DC = Intensity2*(pow(B1, 2) + pow(B2, 2) + pow(B3, 2)) - pow(B3, 2);
		DD = pow(DB, 2) - 4 * DA * DC;

		if(DD < 0.0)
			DD = 0.0;

		D0_neg = (-1 * DB - sqrt(DD)) / (2 * DA);
		D0_pos = (-1 * DB + sqrt(DD)) / (2 * DA);
		if(D0_neg > D0_pos)
		{
			D0_tmp = D0_pos;
			D0_pos = D0_neg;
			D0_neg = D0_tmp;
		}

		if(D0_neg > max_D1_D2)
			(*cur_Psp_FMM).estDepthMap[row_index_0][col_index_0] = D0_neg;
		else if(D0_pos > max_D1_D2)
			(*cur_Psp_FMM).estDepthMap[row_index_0][col_index_0] = D0_pos;
		else if(D0_neg > min_D1_D2 && PERSPECTIVE_FMM_HEURISTIC)
			(*cur_Psp_FMM).estDepthMap[row_index_0][col_index_0] = D0_neg;
		else if(D0_pos > min_D1_D2 && PERSPECTIVE_FMM_HEURISTIC)
			(*cur_Psp_FMM).estDepthMap[row_index_0][col_index_0] = D0_pos;
		else if(D1 < D2)
			D2 = INF;
		else
			D1 = INF;
	}

	if(D1 == INF)
	{
		//============ U
		mode = 1;

		DA = If_2 + (Intensity2 - 1) * (pow(u0, 2) + pow(v0, 2));
		DB = -2.0 * D2 * (If_2 + (Intensity2 - 1) * (u0*u2 + v0*v2));
		DC = pow(D2, 2) * (If_2 + (Intensity2 - 1) * (pow(u2, 2) + pow(v2, 2)));

		DD = pow(DB, 2) - 4 * DA * DC;
		if(DD < 0.0)
			DD = 0.0;

		D0_neg = (-1 * DB - sqrt(DD)) / (2 * DA);
		D0_pos = (-1 * DB + sqrt(DD)) / (2 * DA);
		if(D0_neg > D0_pos)
		{
			D0_tmp = D0_pos;
			D0_pos = D0_neg;
			D0_neg = D0_tmp;
		}

		if(D0_neg > max_D1_D2)
			(*cur_Psp_FMM).estDepthMap[row_index_0][col_index_0] = D0_neg;
		else if(D0_pos > max_D1_D2)
			(*cur_Psp_FMM).estDepthMap[row_index_0][col_index_0] = D0_pos;
		else if(D0_neg > min_D1_D2 && PERSPECTIVE_FMM_HEURISTIC)
			(*cur_Psp_FMM).estDepthMap[row_index_0][col_index_0] = D0_neg;
		else if(D0_pos > min_D1_D2 && PERSPECTIVE_FMM_HEURISTIC)
			(*cur_Psp_FMM).estDepthMap[row_index_0][col_index_0] = D0_pos;
	}
	
	if(D2 == INF)
	{
		//============ V
		mode = 2;

		DA = If_2 + (Intensity2 - 1) * (pow(u0, 2) + pow(v0, 2));
		DB = -2.0 * D1 * (If_2 + (Intensity2 - 1) * (u0*u1 + v0*v1));
		DC = pow(D1, 2) * (If_2 + (Intensity2 - 1) * (pow(u1, 2) + pow(v1, 2)));

		DD = pow(DB, 2) - 4 * DA * DC;
		if(DD < 0.0)
			DD = 0.0;

		D0_neg = (-1 * DB - sqrt(DD)) / (2 * DA);
		D0_pos = (-1 * DB + sqrt(DD)) / (2 * DA);
		if(D0_neg > D0_pos)
		{
			D0_tmp = D0_pos;
			D0_pos = D0_neg;
			D0_neg = D0_tmp;
		}

		if(D0_neg > max_D1_D2)
			(*cur_Psp_FMM).estDepthMap[row_index_0][col_index_0] = D0_neg;
		else if(D0_pos > max_D1_D2)
			(*cur_Psp_FMM).estDepthMap[row_index_0][col_index_0] = D0_pos;
		else if(D0_neg > min_D1_D2 && PERSPECTIVE_FMM_HEURISTIC)
			(*cur_Psp_FMM).estDepthMap[row_index_0][col_index_0] = D0_neg;
		else if(D0_pos > min_D1_D2 && PERSPECTIVE_FMM_HEURISTIC)
			(*cur_Psp_FMM).estDepthMap[row_index_0][col_index_0] = D0_pos;
	}

	return;
}

